# Copyright 2022 Mattia Giambirtone
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import stats
import times
import random
import strformat
import ../src/nimdeque


when isMainModule:
    const size = 1500
    const benchSize = 500000

    echo &"Running tests with queue of size {size}"
    var deque = newLinkedDeque[int]()
    var testStart = cpuTime()
    echo &"\t- Checking add()"
    for i in countup(0, size - 1, 1):
        deque.add(i)
    doAssert deque.len() == size

    echo "\t- Checking iteration"
    for i in countup(0, size - 1, 1):
        doAssert deque[i] == i
    for i, e in deque:
        doAssert i == e
    var j = 0
    for e in deque.reversed():
        doAssert size - j - 1 == e
        inc(j)
    for i, e in deque.reversedPairs():
        doAssert size - i - 1 == e

    echo "\t- Checking contains()"
    for i in countup(0, size - 1, 1):
        doAssert i in deque
    doAssert 48574857 notin deque
    doAssert -0xfffffff notin deque
    echo "\t- Checking pop(0)"
    doAssert deque.pop() == 0
    doAssert deque.len() == size - 1
    doAssert deque[0] == 1

    echo "\t- Checking pop(^1)"
    doAssert deque.pop(deque.high()) == size - 1
    doAssert deque.len() == size - 2
    doAssert deque[deque.high()] == size - 2

    echo "\t- Re-checking values"
    for i in countup(0, size - 3, 1):
        doAssert deque[i] == i + 1

    echo "\t- Checking addLeft()"
    deque.addLeft(0)
    doAssert deque.len() == size - 1
    doAssert deque[0] == 0

    echo "\t- Re-checking values"
    for i in countup(0, size - 2, 1):
        doAssert deque[i] == i

    echo "\t- Checking insert(3)"
    var oldLen = deque.len()
    deque.insert(3, 69420)
    doAssert oldLen + 1 == deque.len()
    doAssert deque.pop(3) == 69420
    doAssert deque.len() == oldLen

    echo &"\t- Checking insert({size - 2})"
    oldLen = deque.len()
    deque.insert(size - 2, 0x42362)
    doAssert oldLen + 1 == deque.len()
    doAssert deque.pop(size - 1) == 0x42362
    doAssert deque.len() == oldLen

    echo &"\t- Checking insert({size div 2})"
    oldLen = deque.len()
    deque.insert(size div 2, 0xf7102)
    doAssert oldLen + 1 == deque.len()
    doAssert deque.pop(size div 2) == 0xf7102
    doAssert deque.len() == oldLen

    randomize()
    let idx = rand(size - 1)
    echo &"\t- Checking insert({idx})"
    oldLen = deque.len()
    deque.insert(size - 2, idx)
    doAssert oldLen + 1 == deque.len()
    doAssert deque.pop(size - 1) == idx
    doAssert deque.len() == oldLen

    echo "\t- Checking backwards indeces"
    for i in countdown(deque.high(), 1):
        doAssert deque[^i] == deque[deque.len() - i]
    deque.add(deque.pop(^1))
    doAssert deque[deque.high()] == deque[^1]

    echo &"\t- Checking queue with maxSize {size div 2}"
    var queue = newLinkedDeque[int](size div 2)
    for i in countup(0, (size div 2) - 1):
        queue.add(i)
    doAssert queue.len() == size div 2
    var temp = queue[0]
    queue.add((size div 2) + 1)
    doAssert queue.len() == size div 2
    doAssert queue[^1] == (size div 2) + 1
    doAssert queue[0] == temp + 1
    queue.addLeft(0)
    doAssert queue.len() == size div 2
    doAssert queue[0] == 0
    doAssert queue[^1] == (size div 2) - 1

    echo "\t- Testing extend()"
    var old = deque.len()
    var s = @[1, 2, 3, 4, 5, 6]
    deque.extend(s)
    for i in countup(size + 5, deque.high()):
        doAssert deque[i] == s[i]
    doAssert old + len(s) == deque.len()

    echo "\t- Testing extendLeft()"
    old = deque.len()
    deque.extendLeft(s)
    for i in countup(0, 5):
        doAssert deque[i] == s[5 - i]

    echo "\t- Testing clear()"
    deque.clear()
    doAssert deque.len() == 0
    doAssertRaises(IndexDefect, echo deque[0])

    echo "\t- Testing equality"
    var 
        a = newLinkedDeque[int]()
        b = newLinkedDeque[int]()
    for i in countup(0, size - 1):
        a.add(i)
    # I could add to both a and b together
    # but ¯\_(ツ)_/¯ who cares
    b.extend(a)
    doAssert a == b
    echo &"Tests completed in {cpuTime() - testStart} seconds"

    ## End of tests, start of benchmark
    echo &"\nRunning benchmarks with queue size of {benchSize} against a seq"
    var q = newLinkedDeque[int]()
    var q2: seq[int] = @[]
    var t: seq[float] = @[]
    var tmp: float
    var st: RunningStat
    var benchStart = cpuTime()
    var start = cpuTime()

    echo &"  Benchmarking LinkedDeque.add()"
    for i in countup(0, benchSize - 1):
        tmp = cpuTime()
        q.add(i)
        t.add(cpuTime() - tmp)
    st.push(t)
    echo &"""
    - Done in {cpuTime() - start} seconds. Results (in seconds):
        - min: {st.min}
        - max: {st.max}
        - avg: {st.mean()}
        - stdev: {st.standardDeviation()}"""

    st.clear()
    t = @[]
    start = cpuTime()
    echo &"  Benchmarking seq.add()"
    for i in countup(0, benchSize - 1):
        tmp = cpuTime()
        q2.add(i)
        t.add(cpuTime() - tmp)
    st.push(t)
    echo &"""
    - Done in {cpuTime() - start} seconds. Results (in seconds):
        - min: {st.min}
        - max: {st.max}
        - avg: {st.mean()}
        - stdev: {st.standardDeviation()}"""

    st.clear()
    t = @[]
    start = cpuTime()
    echo &"  Benchmarking LinkedDeque.pop(0)"
    for i in countup(0, size * 10 - 1):
        tmp = cpuTime()
        discard q.pop()
        t.add(cpuTime() - tmp)
    st.push(t)
    echo &"""
    - Done in {cpuTime() - start} seconds. Results (in seconds):
       - min: {st.min}
       - max: {st.max}
       - avg: {st.mean()}
       - stdev: {st.standardDeviation()}"""

    st.clear()
    t = @[]
    start = cpuTime()
    echo &"  Benchmarking seq.delete(0)"
    for i in countup(0, size * 10 - 1):
        tmp = cpuTime()
        q2.delete(0)
        t.add(cpuTime() - tmp)
    st.push(t)
    echo &"""
    - Done in {cpuTime() - start} seconds. Results (in seconds):
       - min: {st.min}
       - max: {st.max}
       - avg: {st.mean()}
       - stdev: {st.standardDeviation()}"""

    st.clear()
    t = @[]
    start = cpuTime()
    q2 = @[]
    echo &"  Benchmarking LinkedDeque.addLeft()"
    for i in countup(0, benchSize - 1):
        tmp = cpuTime()
        q.addLeft(i)
        t.add(cpuTime() - tmp)
    st.push(t)
    echo &"""
    - Done in {cpuTime() - start} seconds. Results (in seconds):
       - min: {st.min}
       - max: {st.max}
       - avg: {st.mean()}
       - stdev: {st.standardDeviation()}"""

    st.clear()
    t = @[]
    start = cpuTime()
    echo &"  Benchmarking seq.insert(0)"
    for i in countup(0, benchSize - 1):
        tmp = cpuTime()
        q2.insert(i, 0)
        t.add(cpuTime() - tmp)
    st.push(t)
    echo &"""
    - Done in {cpuTime() - start} seconds. Results (in seconds):
       - min: {st.min}
       - max: {st.max}
       - avg: {st.mean()}
       - stdev: {st.standardDeviation()}"""

    st.clear()
    t = @[]
    start = cpuTime()
    echo "  Benchmarking random access for LinkedDeque (10000 times)"
    for i in countup(0, 10000):
        tmp = cpuTime()
        discard q[rand(benchSize - 1)]
        t.add(cpuTime() - tmp)
    st.push(t)
    echo &"""
    - Done in {cpuTime() - start} seconds. Results (in seconds):
       - min: {st.min}
       - max: {st.max}
       - avg: {st.mean()}
       - stdev: {st.standardDeviation()}"""

    st.clear()
    t = @[]
    start = cpuTime()
    echo "  Benchmarking random access for seq (10000 times)"
    for i in countup(0, 10000):
        tmp = cpuTime()
        discard q2[rand(benchSize - 1)]
        t.add(cpuTime() - tmp)
    st.push(t)
    echo &"""
    - Done in {cpuTime() - start} seconds. Results (in seconds):
       - min: {st.min}
       - max: {st.max}
       - avg: {st.mean()}
       - stdev: {st.standardDeviation()}"""

    st.clear()
    t = @[]
    t = @[]
    start = cpuTime()
    echo &"  Benchmarking LinkedDeque.pop(^1)"
    for i in countup(0, size * 10 - 1):
        tmp = cpuTime()
        discard q.pop(^1)
        t.add(cpuTime() - tmp)
    st.push(t)
    echo &"""
    - Done in {cpuTime() - start} seconds. Results (in seconds):
       - min: {st.min}
       - max: {st.max}
       - avg: {st.mean()}
       - stdev: {st.standardDeviation()}"""

    st.clear()
    t = @[]
    start = cpuTime()
    echo &"  Benchmarking seq.pop()"
    for i in countup(0, size * 10 - 1):
        tmp = cpuTime()
        discard q2.pop()
        t.add(cpuTime() - tmp)
    st.push(t)
    echo &"""
    - Done in {cpuTime() - start} seconds. Results (in seconds):
       - min: {st.min}
       - max: {st.max}
       - avg: {st.mean()}
       - stdev: {st.standardDeviation()}"""

    st.clear()
    t = @[]
    start = cpuTime()
    q2 = @[]

    echo &"Total benchmark time: {cpuTime() - benchStart}"
    echo &"Total execution time: {cpuTime() - testStart}"